/*
*
*  Copyright (c) 2005 Michael Stock. All Rights Reserved.
*
*  This file is part of Stirf.
*
*  Stirf is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  Stirf is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this software; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// NOTES ON TIMES/DATES
// GetCurrentEvent returns a structure with local times.
// So do the Hdd access functions, eg: TYPE_File.hour and .min are in local time.
// GetCurrentTime returns local time, not GMT.
// rec file headers have GMT time in them for event/EPG start and end times and dates.

// check status every however-many minutes (in 10ms increments, so for example 60,000 = 10 minutes)
#define CHECK_FREQ  6000
// number of bytes in each event entry in a .epg file
#define EVENT_ENTRY_LEN 1303 // Must be more than EPG_FILE_HEADER
// number of bytes in each section of the event entry, for writing into rec files
#define EVENT_SECTION_1_LEN 14
#define EVENT_SECTION_2_LEN 259
#define EVENT_SECTION_3_LEN 1030
// number of bytes in header of each .epg file
#define EPG_FILE_HEADER 150

static dword lastTickCheck = 0;
static byte minDuration = 3;
static bool updateFailed = FALSE;

typedef struct {
	bool wasRecording; // relates to "the last time we checked"
	char recFileName[ TS_FILE_NAME_SIZE ];
	char EPGfileName[ TS_FILE_NAME_SIZE ];
	word lastEventIDadded;
	byte numEventsAdded;
} SLOT_DATA;

static SLOT_DATA slotZeroStateData, slotOneStateData;

// These time/date functions written by Andy Cullen (?), via Peter Gillespie. Thanks guys.
//------------------------------ NumericDateTime --------------------------------------
//One day span = 86400 so you can work out how long ago something was in a jiffy.
long NumericDateTime( word wDate, byte btHour, byte btMin, byte btSec ) {
	word	wBaseDate; 
	long	lTime;

	wBaseDate = TAP_MakeMjd( 2000, 01, 01 );

	lTime = ( ( wDate - wBaseDate) * 86400 );
	lTime = lTime + ( btHour * 3600 ) + ( btMin * 60 ) + btSec;

	return lTime;
}
//------------------------------ CurrentTime --------------------------------------
//
long	CurrentTime()
{
	word	wDate; 
	byte	btHour, btMin, btSec;

	TAP_GetTime( &wDate, &btHour, &btMin, &btSec );
	return  NumericDateTime( wDate, btHour, btMin, btSec );
}

//------------------------ Replacement function for TAP_Hdd_Delete ----------------
// TAP_Hdd_Delete("Mr. & Mrs. Smith (Mr. & Mrs. Smith).epg") would freeze the toppy
word My_Hdd_Delete(char * s) {
	int i = 0;
	char *dotpos;
	dotpos = strchr(s, '.');
	if (dotpos != NULL)
		i = strlen(dotpos);
	else
		i = 0;
	if (i > 30) {
		TAP_Hdd_Rename(s, "~eurostirf.tmp~");
		return TAP_Hdd_Delete("~eurostirf.tmp~");
	} else
	{
		return TAP_Hdd_Delete(s);
	}
}
//-------------------------------------------------------------------------------------
void resetSlotData (SLOT_DATA *data) {
	data->wasRecording = FALSE;
	memset(data->recFileName, 0, sizeof data->recFileName);
	memset(data->EPGfileName, 0, sizeof data->recFileName);
	data->lastEventIDadded = 0;
	data->numEventsAdded = 0;
}

// Updates the stored "last update" time in the epg file
bool touchEPGfile(SLOT_DATA *data, TYPE_RecInfo *recData) {

	TYPE_File *file;
	char buffer[8];
	// date/time variables
	word mjd; 
	byte hour, min, sec;

	if ( data->numEventsAdded == 0) {
		return FALSE;
	}
	
	memset(buffer, 0, sizeof buffer);
	
    ChangeDirRoot();
 	TAP_Hdd_ChangeDir( "ProgramFiles" );
	TAP_Hdd_ChangeDir( "Stirf" ) ;
	
	if (file = TAP_Hdd_Fopen(data->EPGfileName)) {
		// Write new header for file
		TAP_Hdd_Fseek(file, 1, SEEK_SET);
		TAP_GetTime( &mjd, &hour, &min, &sec );
		buffer[0] = mjd >> 8;
		buffer[1] = mjd & 255;
		buffer[2] = hour;
		buffer[3] = min;
		// Update the currently predicted finish time for the recording
		buffer[4] = (recData->endTime) >> 24; // first byte of end time;s mjd
		buffer[5] = ((recData->endTime) >> 16) & 255;
		buffer[6] = ((recData->endTime) >> 8) & 255; // hour of end time, from 4-byte endTime
		buffer[7] = (recData->endTime) & 255; // minute of end time, from 4-byte endTime
		TAP_Hdd_Fwrite(buffer, 8, 1, file);
		TAP_Hdd_Fclose(file);
	} else {
		// error opening file
		return FALSE;
	}
	
	return TRUE;
}

// Writes data to EPGFilename.
// Returns TRUE on success.
bool addEvent(TYPE_TapEvent *event, SLOT_DATA *data, TYPE_RecInfo *recData) {
	
	TYPE_File *file;
	char *extInfo = NULL;
	char eventBuff[EVENT_ENTRY_LEN];
	word extInfoLen = 0;
	bool useEventDesc = FALSE;
	// date/time variables
	word mjd; 
	byte hour, min, sec;

	// No point adding too many events
	if ((data->numEventsAdded) >= NUM_EVENT_DATA) return FALSE;
	
	memset(eventBuff, 0, EVENT_ENTRY_LEN);
	
	// Doesn't work: TAP_Hdd_ChangeDir( "//" );
    ChangeDirRoot();
 	TAP_Hdd_ChangeDir( "ProgramFiles" );
	TAP_Hdd_ChangeDir( "Stirf" ) ;
	
	// If necessary: create file
	if ( data->numEventsAdded == 0) {
		My_Hdd_Delete(data->EPGfileName); // just in case an old one is still there
		TAP_Hdd_Create(data->EPGfileName, ATTR_NORMAL);
	}
	
	if (file = TAP_Hdd_Fopen(data->EPGfileName)) {
		// Write new header for file
		TAP_Hdd_Fseek(file, 0, SEEK_SET);
		eventBuff[0] = data->numEventsAdded + 1;
		TAP_GetTime( &mjd, &hour, &min, &sec );
		eventBuff[1] = mjd >> 8;
		eventBuff[2] = mjd & 255;
		eventBuff[3] = hour;
		eventBuff[4] = min;
		// Add the currently predicted finish time for the recording
		eventBuff[5] = (recData->endTime) >> 24; // first byte of end time;s mjd
		eventBuff[6] = ((recData->endTime) >> 16) & 255;
		eventBuff[7] = ((recData->endTime) >> 8) & 255; // hour of end time, from 4-byte endTime
		eventBuff[8] = (recData->endTime) & 255; // minute of end time, from 4-byte endTime
		TAP_Hdd_Fwrite(eventBuff, EPG_FILE_HEADER, 1, file);		
		memset(eventBuff, 0, EPG_FILE_HEADER);
		
		// Construct event info buffer. Doing this all in binary ensures we can write from here straight to the .rec file later
		eventBuff[0] =  (event->duration) >> 8; // Event duration - hours
		eventBuff[1] =  (event->duration) & 255; // Event duration - minutes
		eventBuff[4] =  (event->evtId) >> 8; // Event ID - 4 bytes in header but only a 2-byte word in TYPE_TapEvent!!
		eventBuff[5] =  (event->evtId) & 255; // Event ID - 4 bytes in header but only a 2-byte word in TYPE_TapEvent!!

		// Note we do not store GMT times here, even though they need to be in the end when they're put
		//  into the rec file. But doing it now makes it harder to compare against local times later
		//  when we want to compare event start/stop times against (local) recording times and so on.
		//  The switch to GMT for insertion into the .rec file is done later.
		eventBuff[6] =  (event->startTime) >> 24; // MJD is two bytes, from the first two bytes of 4-byte startTime
		eventBuff[7] =  ((event->startTime) >> 16) & 255;
		eventBuff[8] =  ((event->startTime) >> 8) & 255; // hour of start time, from 4-byte startTime
		eventBuff[9] =  (event->startTime) & 255; // minute of start time, from 4-byte startTime
		eventBuff[10] = (event->endTime) >> 24; // MJD is two bytes, from the first two bytes of 4-byte endTime
		eventBuff[11] = ((event->endTime) >> 16) & 255;
		eventBuff[12] = ((event->endTime) >> 8) & 255; // hour of start time, from 4-byte endTime
		eventBuff[13] = (event->endTime) & 255; // minute of start time, from 4-byte endTime
		eventBuff[14] = strlen(event->eventName);
		eventBuff[15] = event->parentalRating;
		strcpy(eventBuff + 16, event->eventName); // write the show title
		strcpy(eventBuff + 16 + eventBuff[14], event->description); // write the show subtitle/description
		// Get extended info, if any
		extInfo = TAP_EPG_GetExtInfo( event );
		if (extInfo != NULL) {
			extInfoLen = strlen(extInfo);
			// stop over-run in eventBuff by limiting extended info length to that handled by a .rec file
			if (extInfoLen >= 1022) {
				extInfo[1022] = '\0';
				extInfoLen = 1022;
			} else if ( (extInfoLen == 0) && ( strlen(event->description) > 0 ) ) {
				extInfoLen = strlen(event->description);
				useEventDesc = TRUE;
			}
			//if ( (extInfoLen == 0) && (hardware == HW_Procaster))  //then maybe try to split the description...
			eventBuff[273] = extInfoLen >> 8;
			eventBuff[274] = extInfoLen & 255;
			if (useEventDesc) {
				strcpy(eventBuff+279,event->description); // use the description instead, as no extended info was found
			} else {
				strcpy(eventBuff+279,extInfo);
			}
			TAP_MemFree( extInfo );
			eventBuff[277] = (event->evtId) >> 8; // Event ID - 4 bytes in header but only a 2-byte word in TYPE_TapEvent!!
			eventBuff[278] = (event->evtId) & 255; // Event ID - 4 bytes in header but only a 2-byte word in TYPE_TapEvent!!
		} else { // try to use description instead, as no extended info was found
			extInfoLen = strlen(event->description);
			if (extInfoLen > 0) {
				eventBuff[273] = extInfoLen >> 8;
				eventBuff[274] = extInfoLen & 255;
				strcpy(eventBuff+279,event->description);
				eventBuff[277] = (event->evtId) >> 8;
				eventBuff[278] = (event->evtId) & 255;
			}
		}
		// Write EPG data to file
		TAP_Hdd_Fseek(file, EPG_FILE_HEADER + (data->numEventsAdded)*EVENT_ENTRY_LEN, SEEK_SET);
		// Write event data
		TAP_Hdd_Fwrite(eventBuff, EVENT_ENTRY_LEN, 1, file);
		// Finish up and close file
		TAP_Hdd_Fclose(file);
		data->numEventsAdded = data->numEventsAdded+1;
		data->lastEventIDadded = event->evtId;
	} else {
		// error opening file
		return FALSE;
	}
	
	return TRUE;
}

// Determines which event is the best one to use as the main EPG entry for the file.
// Returns an array index style number: greater than or equal to zero, and less than numEventsInFile.
byte determineMainEventIndex(byte *eventData, TYPE_File *recFile, byte numEventsInFile, long recFileStartTime, long recFileEndTime) {
	
	byte mainEvent = 1;
	byte *sectionData = NULL;
	long showStartTimes[NUM_EVENT_DATA];
	long showEndTimes[NUM_EVENT_DATA];
	float fileLengthSec = 0;
	float fractionsOfRec[NUM_EVENT_DATA];
	float lastNameFraction = 0.0;
	byte nameResult = 0;
	byte i;
	byte maxFractionIndex = 0;
	word numMins = 0;
	
	fileLengthSec = recFileEndTime - recFileStartTime;
	
	// Determine the fraction of the rec file for which each show was playing, and check for shows that match the
	//  file name.
	for (i = 0; i < numEventsInFile; i++ ) {
		sectionData = eventData + EVENT_ENTRY_LEN*i;
		showStartTimes[i] = NumericDateTime( ( *(sectionData + 6 ))*256 + ( *(sectionData + 7 )), *(sectionData + 8 ), *(sectionData + 9 ), 0 );
		showEndTimes[i] =   NumericDateTime( ( *(sectionData + 10))*256 + ( *(sectionData + 11)), *(sectionData + 12), *(sectionData + 13), 0 );
		if (recEPGfixMode >= Mode_Fixed10) { // if user has asked for EPG data from a fixed %age through the recording
			if (  showEndTimes[i] > (recFileStartTime + fileLengthSec*(recEPGfixMode-Mode_Fixed10+1)/10) ) {
				// Then return the first one satisfying this %age condition
				return i;
			}
		} 
		fractionsOfRec[i] = (min(recFileEndTime,showEndTimes[i])-max(showStartTimes[i],recFileStartTime))/fileLengthSec;
		// Try to match name of show with name of rec file
		if ( (fractionsOfRec[i] > (max(lastNameFraction,0.25))) && (strncmp(recFile->name,sectionData+16,*(sectionData+14)) == 0) ) {
			nameResult = i + 1;
			lastNameFraction = fractionsOfRec[i];
		}
		if (fractionsOfRec[i] > fractionsOfRec[maxFractionIndex]) {
			maxFractionIndex = i;
		}
	}
	
	// If we matched a show name with the filename then go straight to that
	if (nameResult > 0) return (nameResult - 1);
	
	// Otherwise just return the one representing the largest percentage fraction of the rec file
	return maxFractionIndex;
}

// Reads the .epg file, determines which is the best event to use for the .rec file,
// constructs a new header and writes it to the file.
bool applyEPGfile(char *EPGfileName, char *recFileName) {

	//long eventStartTime, recStartTime, recEndTime;
	byte buffer[EVENT_ENTRY_LEN];
	byte *eventData = NULL;
	dword Hsize, freeHsize, availHsize;
	TYPE_File *EPGfile, *recFile;
	byte numEventsInFile = 0;
	byte mainEventIndex = 0; // 0 (not one) is the first entry
	char* titlePtrs[NUM_EVENT_DATA];
	char* extInfoPtrs[NUM_EVENT_DATA];
	byte *tempPtr = NULL;
	byte i = 0;
	int totalMinutes = 0;
	word mjd = 0;
	long recFileEndTime = 0;
	long recFileStartTime = 0;
	bool doSubtitle = FALSE;
	
	if ( strlen(EPGfileName) <= 4 ) return FALSE;
	
	memset(buffer, 0, sizeof buffer);
	memset(titlePtrs, 0, sizeof titlePtrs);
	memset(extInfoPtrs, 0, sizeof extInfoPtrs);
	
	// Open the .epg file
    ChangeDirRoot();
 	TAP_Hdd_ChangeDir( "ProgramFiles" );
	TAP_Hdd_ChangeDir( "Stirf" ) ;
	if (EPGfile = TAP_Hdd_Fopen(EPGfileName)) {
		// Read header from EPG file
		TAP_Hdd_Fseek(EPGfile, 0, SEEK_SET);
		TAP_Hdd_Fread(buffer, EPG_FILE_HEADER, 1, EPGfile);
		numEventsInFile = buffer[0];
		if (numEventsInFile == 0) {
			TAP_Hdd_Fclose(EPGfile);
			return TRUE;
		}
		// Read rec file end time - estimate via last update time plus half of interval
		recFileEndTime = NumericDateTime( ( *(buffer + 1))*256 + ( *(buffer + 2)), *(buffer + 3), *(buffer + 4), 0 ) + CHECK_FREQ/200;
		// Begin preparations for reading the events stored in the .epg file
		TAP_MemInfo(&Hsize, &freeHsize, &availHsize);
		if (availHsize > EVENT_ENTRY_LEN*numEventsInFile) {
			eventData = (byte *)TAP_MemAlloc(EVENT_ENTRY_LEN*numEventsInFile); // this can be several kilobytes
		}
		if (eventData == NULL) {
			TAP_Hdd_Fclose(EPGfile);
			return FALSE;
		}
		// Read the events stored in the .epg file
		memset(eventData, 0, EVENT_ENTRY_LEN*numEventsInFile);
		TAP_Hdd_Fseek(EPGfile, EPG_FILE_HEADER, SEEK_SET);
		TAP_Hdd_Fread(eventData, EVENT_ENTRY_LEN, numEventsInFile, EPGfile);
		TAP_Hdd_Fclose(EPGfile);
		
	    ChangeDirRoot();
	 	TAP_Hdd_ChangeDir( "DataFiles" );
		if (recFile = TAP_Hdd_Fopen(recFileName)) {
			
			recFileStartTime = NumericDateTime(recFile->mjd,recFile->hour,recFile->min,recFile->sec);
			// Cross-check the calculated end time with start time plus calculated file duration. Assume a large gig to
			// minute ratio, get a conservatively small minutes estimate, use it if the "touch file" didn't work (which is
			// undetectable here of course...)
			// minutes = max(5,(min(toFile->size,((ulong64) 1) << 35) >> 26)); // note >> 25 is "real", >> 26 is conservatively small
			recFileEndTime = max(recFileEndTime, recFileStartTime + (max(5,(min(recFile->size,((ulong64) 1) << 35) >> 26)))*60 );
			
			// Set mainEventIndex here. 0 (not 1) means the first entry
			mainEventIndex = determineMainEventIndex(eventData, recFile, numEventsInFile, recFileStartTime, recFileEndTime);
		
			for (i = mainEventIndex; (i < numEventsInFile) && (i < NUM_EVENT_DATA); i++) {
				titlePtrs[i-mainEventIndex] = eventData + EVENT_ENTRY_LEN*i + EVENT_SECTION_1_LEN + 2;
				extInfoPtrs[i-mainEventIndex] = eventData + EVENT_ENTRY_LEN*i + EVENT_SECTION_1_LEN + EVENT_SECTION_2_LEN + 6;
			}
			
			// Construct new .rec file header
			// Write the multi-entry title/subtitle/description text section of what will be the new header into buffer
			memset(buffer, 0, sizeof buffer);
			constructTitleDescriptionArea(buffer, titlePtrs, min(numEventsInFile,NUM_EVENT_DATA) - mainEventIndex, 0);
			// Write buff's new multi-entry title/subt/desc text section over the original section in the new main event
			memcpy(eventData + EVENT_ENTRY_LEN*mainEventIndex + EVENT_SECTION_1_LEN + 2, buffer, SUBT_AREA_SIZE);
			
			// Procaster records the event name length as 1 character longer in the header, as it counts the 0x05
			//  character that it puts at the start. We must therefore do the same, as we replicate the 0x05.
			if (hardwareType==HW_Procaster) {
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + EVENT_SECTION_1_LEN) += 1;
			}
			
			// Write the multi-entry extended info section of what will be the new header into buffer
			memset(buffer, 0, sizeof buffer);
			constructExtendedInfoArea(buffer, extInfoPtrs, min(numEventsInFile,NUM_EVENT_DATA) - mainEventIndex, 0);
			// Write buff's new multi-entry extended info text section over the original section in the new main event
			memcpy(eventData + EVENT_ENTRY_LEN*mainEventIndex + EVENT_SECTION_1_LEN + EVENT_SECTION_2_LEN + 6, buffer, EXT_INFO_AREA_SIZE);
			
			// Fix the main EPG entry's times - shift them from local to GMT time (subtract the offset)
			totalMinutes = (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 8))*60 + (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 9));
			totalMinutes -= 15*GMToffset15;
			if (totalMinutes < 0) {
				// Decrement date by one day
				mjd = (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 6))*256 + (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 7)) - 1;
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 6) = (mjd >> 8);
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 7) = (mjd & 255);
				totalMinutes += 24*60;
			} else if (totalMinutes >= 24*60) {
				// Increment date by one day
				mjd = (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 6))*256 + (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 7)) + 1;
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 6) = (mjd >> 8);
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 7) = (mjd & 255);
				totalMinutes -= 24*60;
			}
			// Fix hours byte
			*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 8) = totalMinutes / 60;
			// Fix minutes byte
			*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 9) = totalMinutes % 60;
			
			// now fix the EPG entry's end time
			totalMinutes = (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 12))*60 + (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 13));
			totalMinutes -= 15*GMToffset15;
			if (totalMinutes < 0) {
				// Decrement date by one day
				mjd = (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 10))*256 + (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 11)) - 1;
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 10) = (mjd >> 8);
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 11) = (mjd & 255);
				totalMinutes += 24*60;
			} else if (totalMinutes >= 24*60) {
				// Increment date by one day
				mjd = (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 10))*256 + (*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 11)) + 1;
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 10) = (mjd >> 8);
				*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 11) = (mjd & 255);
				totalMinutes -= 24*60;
			}
			// Fix hours byte
			*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 12) = totalMinutes / 60;
			// Fix minutes byte
			*(eventData + EVENT_ENTRY_LEN*mainEventIndex + 13) = totalMinutes % 60;
						
			if (recEPGfixMode != Mode_Off) 
			{
				// APPLY HEADER TO FILE
				// Apply each of the three sections
				// First section from Duration to End_Time_Min
				TAP_Hdd_Fseek(recFile, 70+EPGoffset, SEEK_SET);
				TAP_Hdd_Fwrite(eventData + EVENT_ENTRY_LEN*mainEventIndex,                                             EVENT_SECTION_1_LEN, 1, recFile);
				// Second section from Event_Name_Length to Event Name Text
				TAP_Hdd_Fseek(recFile, 85+EPGoffset, SEEK_SET);
				TAP_Hdd_Fwrite(eventData + EVENT_ENTRY_LEN*mainEventIndex + EVENT_SECTION_1_LEN,                       EVENT_SECTION_2_LEN, 1, recFile);
				// Third section from Extended Info Len to Extended Info Text
				TAP_Hdd_Fseek(recFile, 362+EPGoffset, SEEK_SET);
				TAP_Hdd_Fwrite(eventData + EVENT_ENTRY_LEN*mainEventIndex + EVENT_SECTION_1_LEN + EVENT_SECTION_2_LEN, EVENT_SECTION_3_LEN, 1, recFile);
			}
		} else {
			TAP_MemFree(eventData);
			return FALSE;
		}
		// Finish up
		TAP_Hdd_Fclose(recFile);
		// Rename the file if option to do so is turned on, and if conditions met
		i = (*(eventData+EVENT_ENTRY_LEN*mainEventIndex+14)); // title length of correct show
		if ( (autoRenameFiles != Rename_Mode_Off) && (i > 1) ) {
			// If filename matches first show, and first show was not the final show chosen...
			//  OR
			// if the filename didn't contain a + sign and the user has selected filename subtitling and the subtitle is OK
			doSubtitle = ( (autoRenameFiles == Rename_Mode_On_Subtitle) && ( strchr(recFileName,'+') == NULL ) && ( *(eventData+EVENT_ENTRY_LEN*mainEventIndex+16+i) != '\0' ) && ( *(eventData+EVENT_ENTRY_LEN*mainEventIndex+16+i) != '[' ) );
			// only do Subtitles when title and subtitle are different
			doSubtitle = doSubtitle && 
				strncmp(eventData+EVENT_ENTRY_LEN*mainEventIndex+16,eventData+EVENT_ENTRY_LEN*mainEventIndex+16+i,min(i,*(eventData+EVENT_ENTRY_LEN*mainEventIndex+14+i)));
			if ( ( (strncmp(recFileName,eventData+16,strlen(recFileName)-4) == 0) && (numEventsInFile > 1) && (mainEventIndex > 0) ) || doSubtitle ) {
				memset(buffer,0,TS_FILE_NAME_SIZE);
				// Put the show title into buffer
				strncpy(buffer,eventData+EVENT_ENTRY_LEN*mainEventIndex+16,min( (*(eventData+EVENT_ENTRY_LEN*mainEventIndex+14)),58 )); // from eventData, according to the main show's name's length
				if ( doSubtitle ) {
					// Add the subtitle, being careful with string lengths (API restriction on filename lengths)
					strcat(buffer," - ");
					strncat(buffer,eventData+EVENT_ENTRY_LEN*mainEventIndex+16+i,58 );
					buffer[57] = '\0';
					// Get rid of genre info and trim any trailing spaces etc
					tempPtr = strrchr(buffer,'[');
					if (tempPtr != NULL) {
						*tempPtr = '\0';
					}
					tempPtr = strchr(buffer,'\0');
					tempPtr--;
					while ( ((*tempPtr == ' ')||(*tempPtr == '-')) && (tempPtr > (buffer+2)) ) {
						*tempPtr = '\0';
					}
				}
				strcat(buffer,".rec");
				if (trapIllegalChars) {
					convertIllegalChars(buffer);
					KillCtrChar(buffer);
				}
				TAP_Hdd_Rename(recFileName,buffer);
			}
		}
		TAP_MemFree(eventData);
	} else { // Couldn't open .epg file
		return FALSE;
	}
	
	//eventStartTime = NumericDateTime( (currEvent->startTime)>>16, ((currEvent->startTime) >> 8) & 255, (currEvent->startTime) & 255, 0 );
	//recStartTime =   NumericDateTime( (recData->startTime)>>16,   ((recData->startTime) >> 8) & 255,   (recData->startTime) & 255,   0 );
	//recEndTime =     NumericDateTime( (recData->endTime)>>16,     ((recData->endTime) >> 8) & 255,     (recData->endTime) & 255,     0 );
	
	return TRUE;
}

// Determines if an event should be added to the EPG file
bool shouldAddEvent(TYPE_RecInfo *recData, SLOT_DATA *slotStateData, TYPE_TapEvent *currEvent) {
	
	// check event is long enough, and that we haven't added it before
	return ( ((((currEvent->duration)>>8) >= 1) || (((currEvent->duration)|255) >= minDuration)) && (currEvent->evtId != slotStateData->lastEventIDadded) );
	
}

// Updates the EPG file for the given recording slot with the current
// event, if necessary. Also creates the EPG file if it hasn't been created yet.
void updateEPGfile (TYPE_RecInfo *recData, SLOT_DATA *slotStateData) {
	
	TYPE_TapEvent *currEvent = NULL;

	// check if current event should be added
	if ( currEvent = TAP_GetCurrentEvent(SVC_TYPE_Tv, recData->svcNum) ) {
		if ( (strlen(currEvent->eventName) > 0) && shouldAddEvent(recData, slotStateData, currEvent) ) {
			addEvent(currEvent, slotStateData, recData); // this will create the file if necessary
		} else {
			// maintain the time stored in epg file
			touchEPGfile(slotStateData, recData);
		}
		//TAP_MemFree(currEvent); // Not sure about this. Topfield's examples don't do it, they do it for getEvent but not getCurrentEvent.
	} else {
		touchEPGfile(slotStateData, recData);
	}
}

// Handles the actions required according to the recording state of a slot.
void manageRecordingSlot (TYPE_RecInfo *recData, SLOT_DATA *slotStateData, bool recordingNow) {
	
	if (recordingNow) { // then we can access recData, it shouldn't be NULL
		if ( !(slotStateData->wasRecording) || (strcmp(recData->fileName,slotStateData->recFileName) != 0) ) {
			// then we have a new recording in the slot. Check if another one just finished:
			if ( slotStateData->wasRecording && (strcmp(recData->fileName,slotStateData->recFileName) != 0) ) {
				// then we have just finished a recording in the slot
				if (applyEPGfile(slotStateData->EPGfileName, slotStateData->recFileName)) {
					ChangeDirRoot();
					TAP_Hdd_ChangeDir( "ProgramFiles" );
					TAP_Hdd_ChangeDir( "Stirf" );
					My_Hdd_Delete(slotStateData->EPGfileName);
				} else {
					// The attempted update failed, set marker to try again later
					updateFailed = TRUE;
				}
			}
			// handle the new recording
			resetSlotData(slotStateData);
			strcpy(slotStateData->recFileName, recData->fileName);
			strcpy(slotStateData->EPGfileName,slotStateData->recFileName);
			// Replace ".rec" with ".epg"
			strcpy(strchr(slotStateData->EPGfileName,'\0')-3,"epg");
		} // Finished handling new recordings.
		// Add event info to EPG file if appropriate
		updateEPGfile(recData, slotStateData);
		// update state variable
		slotStateData->wasRecording = TRUE;
	} else { // the slot is not recording now, but was recording previously (otherwise this function wouldn't have been called)
		// NOTE: recData may be null if this is the case
		if (applyEPGfile(slotStateData->EPGfileName, slotStateData->recFileName)) {
			ChangeDirRoot();
			TAP_Hdd_ChangeDir( "ProgramFiles" );
			TAP_Hdd_ChangeDir( "Stirf" );
			My_Hdd_Delete(slotStateData->EPGfileName);
		} else {
			// The attempted update failed, set marker to try again later
			updateFailed = TRUE;
		}
		resetSlotData(slotStateData); // includes setting wasRecording = FALSE; for next time
	}
	
}

// Searches through all .epg files and tries to apply them all. Deletes +3 day old ones too.
void applyAllEPGfiles(void) {
	
	TYPE_File file;
	TYPE_File *openedFile = NULL;
	dword numFilesAndDirs = 0;
	dword numNames = 0;
	int fileIndex, fileIndex2;
	int EPGfilesFound = 0;
//	long *EPGfileTimes = NULL; // API fails to apply correct time/datestamps
	char **EPGfileNames = NULL;
	byte *EPGfileApply = NULL;
	char fileName1[TS_FILE_NAME_SIZE];
	char fileName2[TS_FILE_NAME_SIZE];
	TYPE_RecInfo recInf;
	char buff[4];
	
	fileName1[0] = '\0';
	fileName2[0] = '\0';
	ChangeDirRoot();
 	TAP_Hdd_ChangeDir( "ProgramFiles" );
 	TAP_Hdd_ChangeDir( "Stirf" );
	numFilesAndDirs = TAP_Hdd_FindFirst( &file );
	numNames = numFilesAndDirs;
	// Loop through the .epg files and store their names
	if (numFilesAndDirs > 2) {
//		EPGfileTimes = (long *)TAP_MemAlloc(numNames * (sizeof (long)));
		EPGfileNames = (char **)TAP_MemAlloc(numNames * (sizeof (char*)));
		EPGfileApply = (byte *)TAP_MemAlloc(numNames * (sizeof (byte)));
		if ( (EPGfileNames == NULL) || (EPGfileApply == NULL) ) return; //(EPGfileTimes == NULL) || 
		memset(EPGfileNames, 0, numNames * (sizeof (char*)) );
		// Store names and details of all ".epg" files found
		for (fileIndex = 0; (fileIndex < numFilesAndDirs); fileIndex++) {
			if ( (file.attr == ATTR_NORMAL) && (IsFileEPG(file.name)) ) {
//				EPGfileTimes[EPGfilesFound] = NumericDateTime(file.mjd,file.hour,file.min,file.sec);
				EPGfileNames[EPGfilesFound] = (char *)TAP_MemAlloc(TS_FILE_NAME_SIZE * sizeof(char));
				if (EPGfileNames[EPGfilesFound] == NULL) return;
				strcpy(EPGfileNames[EPGfilesFound], file.name);
				EPGfileApply[EPGfilesFound] = 0;
				EPGfilesFound++;
			}
			if (TAP_Hdd_FindNext( &file ) == 0) {
				fileIndex = numFilesAndDirs; //bail out of the for loop as some problem has happened with the HDD access
			}
		}
		if (EPGfilesFound > 0) {
			// Try to match the .epg files we found with a .rec file
			ChangeDirRoot();
 			TAP_Hdd_ChangeDir( "DataFiles" );
			numFilesAndDirs = TAP_Hdd_FindFirst( &file );
			if (numFilesAndDirs > 0) {
				// Store names of rec files currently recording
				if (TAP_Hdd_GetRecInfo (0, &recInf)) {
					if ( recInf.recType == RECTYPE_Normal) {
						strcpy(fileName1,recInf.fileName);
					}
				}
				if (TAP_Hdd_GetRecInfo (1, &recInf)) {
					if ( recInf.recType == RECTYPE_Normal) {
						strcpy(fileName2,recInf.fileName);
					}
				}
				// Loop through all .rec files
				for (fileIndex = 0; (fileIndex < numFilesAndDirs); fileIndex++) {
					if ( (file.attr == ATTR_TS) && (IsFileRec(file.name)) ) {
						// Search through all epg files we found, looking for a rec file name match (barring .extension)
						//  but don't match with files recording now.
						if ( (strcmp(fileName1, file.name) != 0) && (strcmp(fileName2, file.name) != 0) && (file.name[0] != 0) ) {
							// Loop through .epg files
							for (fileIndex2 = 0; (fileIndex2 < EPGfilesFound); fileIndex2++) {
								// Check if names match (without extension), and check file creation times are OK
								if ( (strncmp(file.name,EPGfileNames[fileIndex2],strlen(file.name)-3) == 0)){// && (EPGfileTimes[fileIndex2] > NumericDateTime(file.mjd,file.hour,file.min,file.sec))  ) {
									// date check in line above disabled due to API datestamping failure (always creates files with 2003-01-01 or something equally stupid)
									EPGfileApply[fileIndex2] = 1; // Set marker for later
									fileIndex2 = EPGfilesFound; // Break out of inner search loop
								}
							}
						}
					}
					if (TAP_Hdd_FindNext( &file ) == 0) {
						fileIndex = numFilesAndDirs; //bail out of the for loop as some problem has happened with the HDD access
					}
				}
				// Now loop through looking for matches (in prev. loop we didn't want to mess up HDD access)
				for (fileIndex2 = 0; (fileIndex2 < EPGfilesFound); fileIndex2++) {
					// Check markers
					if (EPGfileApply[fileIndex2] == 1) {
						strcpy(fileName1,EPGfileNames[fileIndex2]);
						strcpy(strchr(fileName1,'\0')-3,"rec");
						if(applyEPGfile(EPGfileNames[fileIndex2],fileName1)) {
							EPGfileApply[fileIndex2] = 2; // Set marker to delete the epg file
						}
					}
				}
			}
			
			// Delete all used .epg files, and those older than three days
			ChangeDirRoot();
			TAP_Hdd_ChangeDir( "ProgramFiles" );
			TAP_Hdd_ChangeDir( "Stirf" );
			// For each file check if necessary to delete
			for (fileIndex2 = 0; (fileIndex2 < EPGfilesFound); fileIndex2++) {
				// Check markers and dates
				if ((EPGfileApply[fileIndex2] == 2)) { // || ((EPGfileTimes[fileIndex2] + 3*86400) < currTime) ) {
					My_Hdd_Delete(EPGfileNames[fileIndex2]);
				} else {
					// Check if the file is more than 3 days old. Read from inside file, to work around API datestamp bug
					if (openedFile = TAP_Hdd_Fopen(EPGfileNames[fileIndex2])) {
						TAP_Hdd_Fseek(openedFile, 1, SEEK_SET);
						TAP_Hdd_Fread(buff, 4, 1, openedFile);
						TAP_Hdd_Fclose(openedFile);
						// Delete it if it is more than three days old
						if ( (NumericDateTime(buff[0]*256 + buff[1],buff[2],buff[3],0) + 3*86400) < CurrentTime() ) {
							My_Hdd_Delete(EPGfileNames[fileIndex2]);
						}
					}
				}
			}
		} // End of if EPGfilesFound > 0
		// Clean up memory allocated
//		if (EPGfileTimes != NULL) TAP_MemFree(EPGfileTimes);
		for (fileIndex = 0; fileIndex < numNames; fileIndex++) {
			if (EPGfileNames[fileIndex] != NULL) TAP_MemFree(EPGfileNames[fileIndex]);
		}
		if (EPGfileNames != NULL) TAP_MemFree(EPGfileNames);
		if (EPGfileApply != NULL) TAP_MemFree(EPGfileApply);		
	}
	
}

dword recEPGfixEventHandler(void) {
	
	bool slotZeroRecordingNow = FALSE;
	bool slotOneRecordingNow = FALSE;
	TYPE_RecInfo slotZeroRecData, slotOneRecData;
	TYPE_PlayInfo pInfo[2]; // spare one to stop API buffer over-run being a problem
	
	// if event idle...
	// check if it's time to poll the toppy's recording/playback status
	if ( TAP_GetTick() > (lastTickCheck + CHECK_FREQ) ) {
		
		// now we will schedule the next check
		lastTickCheck = TAP_GetTick();
		
		// First check if any old/previous attempts to update failed
		if ( updateFailed ) {
			if ( TAP_Hdd_GetPlayInfo(pInfo) ) {
				if ((pInfo[0].playMode == PLAYMODE_None) || (pInfo[0].playMode == PLAYMODE_TempPlaying) || (pInfo[0].playMode == PLAYMODE_Mp3)) {
					applyAllEPGfiles();
					updateFailed = FALSE; // Only try once
				}
			}
		}
		
		// get recording status
		if (TAP_Hdd_GetRecInfo (0, &slotZeroRecData)) {
			slotZeroRecordingNow = ( (slotZeroRecData.recType == RECTYPE_Normal) && (slotZeroRecData.svcType == SVC_TYPE_Tv) );
		}
		if (TAP_Hdd_GetRecInfo (1, &slotOneRecData)) {
			slotOneRecordingNow = ( (slotOneRecData.recType == RECTYPE_Normal) && (slotOneRecData.svcType == SVC_TYPE_Tv) );
		}
		
		// make the calls to manage the changes in recording status that have been detected, if any
		if ( slotZeroRecordingNow || slotZeroStateData.wasRecording ) {
			manageRecordingSlot( &slotZeroRecData, &slotZeroStateData, slotZeroRecordingNow);
		}
		if ( slotOneRecordingNow || slotOneStateData.wasRecording ) {
			manageRecordingSlot( &slotOneRecData, &slotOneStateData, slotOneRecordingNow);
		}
		
		// Finish off by setting states
		slotZeroStateData.wasRecording = slotZeroRecordingNow;
		slotOneStateData.wasRecording = slotOneRecordingNow;
		
		// And lastly check we haven't stuffed up the current directory if the blue stirf screen is showing
		if (stirfBoxShowing) {
			ChangeDirRoot();
	 		ChangeDir (uGbl.uCurOnScreenuDir.path );
		}
		
	}
	
}

// "stirf.ini" file must be loaded before calling this startup routine.
int recEPGfixMain(void) {
		
	resetSlotData(&slotZeroStateData);
	resetSlotData(&slotOneStateData);
	
	ChangeDirRoot();
 	TAP_Hdd_ChangeDir( "ProgramFiles" );
 	if ( !TAP_Hdd_Exist( "Stirf" ) ) {
	 	TAP_Hdd_Create ( "Stirf", ATTR_FOLDER );
 	}
	// Do a check on startup - are there any .epg files needing either applying to rec file, or deleting?
	applyAllEPGfiles();
	
	// Set timer for starting normal operation
	lastTickCheck = TAP_GetTick();
	
	return 1;
	
}

