/*
*
*  Copyright (c) 2005 Michael Stock. All Rights Reserved.
*
*  This file is part of Stirf.
*
*  Stirf is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  Stirf is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this software; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------constructTitleDescriptionArea(...)-------------------------------------------------
//
// Constructs the title/subtite/description area so you can insert it into a .rec file at offset 87 (91 for 5800).
//
// buff must be allocated before you call this function, with allocated size of at least SUBT_AREA_SIZE bytes
//  (it is imperative to set all of buff to zero - e.g. use memset(...) -  before calling this function).
//
// titleEntries is an array of pointers, each of which points to a string that is a title/subtitle or
//  title/description combination. Each title/subtitle or title/description entry should be the title
//  first with the subtitle or description concatenated after it (without a space), just like they are
//  in rec file entries. eg "The SimpsonsTreehouse of Horror" or "SeinfeldThe Close Talker"
//  or "Some ShowFred does something, Bob does something else, blah blah..."
//
// numPointers is the actual number of pointers in titleEntries (i.e. size of the first array dimension) so that
//  titleEntries[0] through to to titleEntries[numPointers-1] are valid strings (and are *NOT* NULL!).
// 
// mainEntryNum is an array index into titleEntries that signifies which pointer is the one you want as the main EPG
//  entry in the file. This will be copied to buff first, with all the others (if any) put in as spares.
//  Must be LESS THAN numPointers, and at least zero.
//
// NOTE: uses strcpy, therefore the pointers in titleEntries MUST NOT point anywhere into buff.
//
void constructTitleDescriptionArea (char *buff, char **titleEntries, const byte numPointers, const byte mainEntryNum) {
	
	int i;
	char* tempPtr = NULL;
	
	if (numPointers == 0) return;
	
	// Must add in the 0x05 byte for the Procaster
	buff[0] = 5;
	// Over-write the 0x05 if not Procaster
	strcpy( buff + ((hardwareType==HW_Procaster) ? 1:0), titleEntries[mainEntryNum] );				
	tempPtr = buff + strlen(buff);
	if ( numPointers > 1) {
		// save the other data we found
		for (i = 0; i < numPointers; i++) {
			if ( i != mainEntryNum ) { //i.e. don't duplicate entries
				// replicate the Toppy's structure here, with zeros up to a 64 ('@') which is at a multiple of 16
				tempPtr = buff - 8 + 16*(((tempPtr + 10 - buff) / 16) + 1); // was (tempPtr + 10 - buff + 7) - try to squeeze some more in!
				if (tempPtr < (buff + READ_BUFFER_SIZE - 7) ) {
					*tempPtr = '@';
					tempPtr += 7;
					// tempPtr now points to the 'possibly the rating' byte. leave at zero for now,
					// except for the Procaster where it seems to always be 0x05
					if (hardwareType == HW_Procaster) *tempPtr = 5;
					tempPtr += 1; // move to the start of the text field
					*tempPtr = '\0'; // just in case
					strncat(tempPtr, titleEntries[i], READ_BUFFER_SIZE - (tempPtr - buff));
					tempPtr += strlen(tempPtr);
				}
			}
		}
	}
}

// ----------------------------------------constructExtendedInfoArea(...)-------------------------------------------------
//
// Constructs the extended info area so you can insert it into a .rec file at offset 368 (372 for 5800).
//
// buff must be allocated before you call this function, with allocated size of at least EXT_INFO_AREA bytes
//  (it is imperative that you set all of buff to zero - e.g. use memset(...) -  before calling this function).
//
// extInfoEntries is an array of pointers, each of which points to a string that is an extended info entry.
//
// numPointers is the actual number of pointers in extInfoEntries (i.e. size of the first array dimension) so that
//  extInfoEntries[0] through to extInfoEntries[numPointers-1] are valid strings (and are *NOT* NULL!).
// 
// mainEntryNum is an array index into extInfoEntries that signifies which pointer is the one you want as the
// main EPG entry in the file. This will be copied to buff first, with all the others (if any) put in as
// spares. Must be LESS THAN numPointers, and at least zero. Can point to an empty string ("\0").
//
// NOTE: uses strcpy, therefore the pointers in extInfoEntries MUST NOT point anywhere into buff.
//
void constructExtendedInfoArea (char *buff, char **extInfoEntries, const byte numPointers, const byte mainEntryNum) {
	
	int i;
	char* tempPtr = NULL;
	
	if (numPointers == 0) return;
	
	// handle empty strings! including possibly the first one.
	
	// build a new extended data region with the old entries re-ordered
	strcpy( buff, extInfoEntries[mainEntryNum] );
	tempPtr = buff + strlen(buff);
	if ( numPointers > 1) {
		// save the other data we found
		for (i = 0; i < numPointers; i++) {
			if ( i != mainEntryNum ) { //i.e. don't duplicate
				// replicate the Toppy's structure here, with zeros up to a 64 ('@') which is at a multiple of 16
				tempPtr = buff - 8 + 16*(((tempPtr + 3 - buff + 9) / 16) + 1);
				if (tempPtr < (buff + EXT_INFO_AREA_SIZE - 7) ) {
					*tempPtr = '@';
					tempPtr += 7;
					// tempPtr now points to a mystery byte, a bit like the 'possibly the rating' byte in the title area.
					// Leave this byte at zero for now
					//*tempPtr = something?;
					tempPtr += 1; // move to the start of the text field
					strncat(tempPtr, extInfoEntries[i], EXT_INFO_AREA_SIZE - (tempPtr - buff));
					tempPtr += strlen(tempPtr);
				} else {
					return; // no more room!
				}
			}
		}
	}
	
}

